"""This module contains the `generation` decorator and related utilities for tracing."""

from __future__ import annotations

import os
import inspect
from types import MappingProxyType
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    Protocol,
    ParamSpec,
    TypeAlias,
    overload,
)
from functools import wraps
from contextlib import contextmanager
from contextvars import ContextVar
from collections.abc import Callable, Coroutine, Generator

import orjson
from pydantic import BaseModel
from opentelemetry.trace import format_span_id, get_tracer_provider
from opentelemetry.util.types import AttributeValue

from .spans import Span
from ._utils import (
    Closure,
    call_safely,
    fn_is_async,
    inspect_arguments,
    get_qualified_name,
    create_mirascope_middleware,
)
from .sandbox import SandboxRunner, SubprocessSandboxRunner
from .._client import Lilypad, AsyncLilypad
from .exceptions import LilypadValueError, RemoteFunctionError, LilypadNotFoundError
from ._utils.json import to_text, json_dumps, fast_jsonable
from .._exceptions import NotFoundError
from ._utils.client import NETWORK_EXCEPTIONS, get_sync_client, get_async_client
from ._utils.settings import get_settings
from ._utils.functions import get_signature
from ..types.ee.projects import Label, EvaluationType, annotation_create_params
from ._utils.function_cache import (
    get_cached_closure,
    get_function_by_hash_sync,
    get_deployed_function_sync,
    get_function_by_hash_async,
    get_deployed_function_async,
    get_function_by_version_sync,
    get_function_by_version_async,
)
from ..types.projects.functions import FunctionPublic
from ._utils.serializer_registry import SerializerMap

_P = ParamSpec("_P")
_R = TypeVar("_R")
_R_CO = TypeVar("_R_CO", covariant=True)
_T = TypeVar("_T")


TRACE_MODULE_NAME = "lilypad.lib.traces"


def _get_trace_type(function: FunctionPublic | None) -> Literal["trace", "function"]:
    if function:
        return "function"
    return "trace"


class Annotation(BaseModel):
    data: dict[str, Any] | None

    label: Label | None

    reasoning: str | None

    type: EvaluationType | None


class _TraceBase(Generic[_T]):
    """
    Base class for the Trace wrapper.
    """

    def __init__(self, response: _T, span_id: int, function_uuid: str) -> None:
        self.response: _T = response
        self.function_uuid: str = function_uuid
        self.formated_span_id: str = format_span_id(span_id)
        self._flush: bool = False

    def _force_flush(self) -> None:
        tracer = get_tracer_provider()
        if force_flush := getattr(tracer, "force_flush", None):
            force_flush(timeout_millis=5000)
            self._flush = True

    def _create_body(
        self, project_id: str, span_uuid: str, annotation: tuple[Annotation, ...]
    ) -> list[annotation_create_params.Body]:
        return [
            annotation_create_params.Body(
                data=annotation.data,
                function_uuid=self.function_uuid,
                span_uuid=span_uuid,
                label=annotation.label,
                reasoning=annotation.reasoning,
                type=annotation.type,
                project_uuid=project_id,
            )
            for annotation in annotation
        ]


class Trace(_TraceBase[_T]):
    """
    A simple trace wrapper that holds the original function's response and allows annotating the trace.
    """

    def _get_span_uuid(self, client: Lilypad) -> str | None:
        if not self._flush:
            self._force_flush()
        response = client.projects.functions.spans.list(
            project_uuid=get_settings().project_id, function_uuid=self.function_uuid
        )
        for span in response:
            if span.get("span_id") == self.formated_span_id:
                return span["uuid"]
        return None

    def annotate(self, *annotation: Annotation) -> None:
        """
        Annotate the trace with the given annotation.
        """
        settings = get_settings()
        lilypad_client = get_sync_client(api_key=settings.api_key)
        body = self._create_body(settings.project_id, self._get_span_uuid(lilypad_client), annotation)
        lilypad_client.ee.projects.annotations.create(project_uuid=settings.project_id, body=body)

    def assign(self, *email: str) -> None:
        """Assign the trace to a user by email."""
        settings = get_settings()

        lilypad_client = get_sync_client(api_key=settings.api_key)

        lilypad_client.ee.projects.annotations.create(
            project_uuid=settings.project_id,
            body=[
                annotation_create_params.Body(
                    assignee_email=list(email),
                    function_uuid=self.function_uuid,
                    project_uuid=settings.project_id,
                    span_uuid=self._get_span_uuid(lilypad_client),
                )
            ],
        )

    def tag(self, *tags: str) -> None:
        """
        Annotate the trace with the given tags.
        """
        if not tags:
            return None
        tag_list = list(tags)
        settings = get_settings()
        client = get_sync_client(api_key=settings.api_key)
        span_uuid = self._get_span_uuid(client)
        client.projects.spans.update_tags(span_uuid=span_uuid, tags_by_name=tag_list)


class AsyncTrace(_TraceBase[_T]):
    """
    A simple trace wrapper that holds the original function's response and allows annotating the trace.
    """

    async def _get_span_uuid(self, client: AsyncLilypad) -> str | None:
        if not self._flush:
            self._force_flush()
        response = await client.projects.functions.spans.list(
            project_uuid=get_settings().project_id, function_uuid=self.function_uuid
        )
        for span in response:
            if span.get("span_id") == self.formated_span_id:
                return span["uuid"]
        return None

    async def annotate(self, *annotation: Annotation) -> None:
        """
        Annotate the trace with the given annotation.
        """
        settings = get_settings()
        lilypad_client = get_async_client(api_key=settings.api_key)
        body = self._create_body(settings.project_id, await self._get_span_uuid(lilypad_client), annotation)
        await lilypad_client.ee.projects.annotations.create(project_uuid=settings.project_id, body=body)

    async def assign(self, *email: str) -> None:
        """Assign the trace to a user by email."""
        settings = get_settings()
        async_client = get_async_client(api_key=settings.api_key)

        await async_client.ee.projects.annotations.create(
            project_uuid=settings.project_id,
            body=[
                annotation_create_params.Body(
                    assignee_email=list(email),
                    function_uuid=self.function_uuid,
                    project_uuid=settings.project_id,
                    span_uuid=await self._get_span_uuid(async_client),
                )
            ],
        )

    async def tag(self, *tags: str) -> None:
        """
        Annotate the trace with the given tags.
        """
        if not tags:
            return None
        tag_list = list(tags)
        settings = get_settings()
        client = get_async_client(api_key=settings.api_key)
        span_uuid = await self._get_span_uuid(client)
        await client.projects.spans.update_tags(span_uuid=span_uuid, tags_by_name=tag_list)


# Type definitions for decorator registry
FunctionInfo: TypeAlias = tuple[
    str, str, int, str, dict[str, Any]
]  # (file_path, function_name, line_number, module_name, context)
DecoratorRegistry: TypeAlias = dict[str, list[FunctionInfo]]

# Globals for decorator registry
_RECORDING_ENABLED: bool = False
_DECORATOR_REGISTRY: DecoratorRegistry = {}  # Maps decorator names to lists of function info


def enable_recording() -> None:
    """Enable recording of decorated functions."""
    global _RECORDING_ENABLED
    _RECORDING_ENABLED = True


def disable_recording() -> None:
    """Disable recording of decorated functions."""
    global _RECORDING_ENABLED
    _RECORDING_ENABLED = False


def clear_registry() -> None:
    """Clear the registry of decorated functions."""
    global _DECORATOR_REGISTRY
    _DECORATOR_REGISTRY = {}


DecoratorArgs: TypeAlias = dict[str, Any]
FunctionInfo: TypeAlias = tuple[str, str, int, str, DecoratorArgs]


def _register_decorated_function(
    decorator_name: str, fn: Callable[..., Any], function_name: str, context: dict[str, Any] | None = None
) -> None:
    """Register a function that has been decorated.

    Args:
        decorator_name: The name of the decorator
        fn: The decorated function
        function_name: The name of the function
        context: Optional context information to store with the function
    """
    if not _RECORDING_ENABLED:
        return

    try:
        # Get function information
        file_path: str = inspect.getfile(fn)
        abs_path: str = os.path.abspath(file_path)
        lineno: int = inspect.getsourcelines(fn)[1]
        module_name: str = fn.__module__

        # Add to registry
        if decorator_name not in _DECORATOR_REGISTRY:
            _DECORATOR_REGISTRY[decorator_name] = []

        # Store (file_path, function_name, line_number, module_name)
        _DECORATOR_REGISTRY[decorator_name].append((abs_path, function_name, lineno, module_name, context))
    except (TypeError, OSError):
        # Handle cases where inspect might fail (e.g., built-in functions)
        pass


def get_decorated_functions(decorator_name: str | None = None) -> DecoratorRegistry:
    """Get information about registered decorated functions.

    Args:
        decorator_name: Optional name of decorator to filter by

    Returns:
        Dictionary mapping decorator names to lists of function information tuples
    """
    if decorator_name:
        return {decorator_name: _DECORATOR_REGISTRY.get(decorator_name, [])}
    return _DECORATOR_REGISTRY.copy()


class SyncVersionedFunction(Protocol[_P, _R_CO]):
    """Protocol for the `VersionedFunction` decorator return type."""

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R_CO:
        """Protocol for the `VersionFunction` decorator return type."""
        ...

    def version(
        self,
        forced_version: int,
        sandbox_runner: SandboxRunner | None = None,
    ) -> Callable[_P, _R_CO]:
        """Protocol for the `VersionFunction` decorator return type."""
        ...

    def remote(
        self,
        sandbox_runner: SandboxRunner | None = None,
    ) -> _R_CO:
        """Protocol for the `VersionFunction` decorator return type."""
        ...


class AsyncVersionedFunction(Protocol[_P, _R_CO]):
    """Protocol for the `VersionFunction` decorator return type."""

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> Coroutine[Any, Any, _R_CO]:
        """Protocol for the `VersionFunction` decorator return type."""
        ...

    def version(
        self,
        forced_version: int,
        sandbox_runner: SandboxRunner | None = None,
    ) -> Coroutine[Any, Any, Callable[_P, _R_CO]]:
        """Protocol for the `VersionFunction` decorator return type."""
        ...

    def remote(
        self,
        sandbox_runner: SandboxRunner | None = None,
    ) -> Coroutine[Any, Any, _R_CO]:
        """Protocol for the `VersionFunction` decorator return type."""
        ...


class TraceDecoratedFunctionWithContext(Protocol[_P, _R]):
    """Protocol for the `VersioningDecorator` decorator return type."""

    def __call__(self, trace_ctx: Span, *args: _P.args, **kwargs: _P.kwargs) -> _R: ...


class TraceDecorator(Protocol):
    """Protocol for the `VersioningDecorator` decorator return type."""

    @overload
    def __call__(
        self, fn: TraceDecoratedFunctionWithContext[_P, Coroutine[Any, Any, _R]]
    ) -> Callable[_P, Coroutine[Any, Any, _R]]: ...

    @overload
    def __call__(self, fn: TraceDecoratedFunctionWithContext[_P, _R]) -> Callable[_P, _R]: ...

    @overload
    def __call__(self, fn: Callable[_P, Coroutine[Any, Any, _R]]) -> Callable[_P, Coroutine[Any, Any, _R]]: ...

    @overload
    def __call__(self, fn: Callable[_P, _R]) -> Callable[_P, _R]: ...

    def __call__(
        self,
        fn: TraceDecoratedFunctionWithContext[_P, Coroutine[Any, Any, _R]]
        | TraceDecoratedFunctionWithContext[_P, _R]
        | Callable[_P, _R]
        | Callable[_P, Coroutine[Any, Any, _R]],
    ) -> Callable[_P, _R] | Callable[_P, Coroutine[Any, Any, _R]]:
        """Protocol `call` definition for `VersioningDecorator` decorator return type."""
        ...


class WrappedTraceDecorator(Protocol):
    """Protocol for the `WrappedTraceDecorator` decorator return type."""

    @overload
    def __call__(
        self, fn: TraceDecoratedFunctionWithContext[_P, Coroutine[Any, Any, _R]]
    ) -> Callable[_P, Coroutine[Any, Any, Trace[_R]]]: ...

    @overload
    def __call__(self, fn: TraceDecoratedFunctionWithContext[_P, _R]) -> Callable[_P, Trace[_R]]: ...

    @overload
    def __call__(self, fn: Callable[_P, Coroutine[Any, Any, _R]]) -> Callable[_P, Coroutine[Any, Any, Trace[_R]]]: ...

    @overload
    def __call__(self, fn: Callable[_P, _R]) -> Callable[_P, Trace[_R]]: ...

    def __call__(
        self,
        fn: TraceDecoratedFunctionWithContext[_P, Coroutine[Any, Any, _R]]
        | TraceDecoratedFunctionWithContext[_P, _R]
        | Callable[_P, _R]
        | Callable[_P, Coroutine[Any, Any, _R]],
    ) -> Callable[_P, Trace[_R]] | Callable[_P, Coroutine[Any, Any, Trace[_R]]]:
        """Protocol `call` definition for `VersioningDecorator` decorator return type."""
        ...


class VersionedFunctionTraceDecorator(Protocol):
    """Protocol for the `VersionedFunction` decorator return type."""

    @overload
    def __call__(
        self, fn: TraceDecoratedFunctionWithContext[_P, Coroutine[Any, Any, _R]]
    ) -> AsyncVersionedFunction[_P, _R]: ...

    @overload
    def __call__(self, fn: TraceDecoratedFunctionWithContext[_P, _R]) -> SyncVersionedFunction[_P, _R]: ...

    @overload
    def __call__(self, fn: Callable[_P, Coroutine[Any, Any, _R]]) -> AsyncVersionedFunction[_P, _R]: ...

    @overload
    def __call__(self, fn: Callable[_P, _R]) -> SyncVersionedFunction[_P, _R]: ...

    def __call__(
        self,
        fn: TraceDecoratedFunctionWithContext[_P, Coroutine[Any, Any, _R]]
        | TraceDecoratedFunctionWithContext[_P, _R]
        | Callable[_P, _R]
        | Callable[_P, Coroutine[Any, Any, _R]],
    ) -> AsyncVersionedFunction[_P, _R] | SyncVersionedFunction[_P, _R]:
        """Protocol `call` definition for `VersionedFunction` decorator return type."""
        ...


class WrappedVersionedFunctionTraceDecorator(Protocol):
    """Protocol for the `WrappedVersionedFunctionTraceDecorator` decorator return type."""

    @overload
    def __call__(
        self, fn: TraceDecoratedFunctionWithContext[_P, Coroutine[Any, Any, _R]]
    ) -> AsyncVersionedFunction[_P, Trace[_R]]: ...

    @overload
    def __call__(self, fn: TraceDecoratedFunctionWithContext[_P, _R]) -> SyncVersionedFunction[_P, Trace[_R]]: ...

    @overload
    def __call__(self, fn: Callable[_P, Coroutine[Any, Any, _R]]) -> AsyncVersionedFunction[_P, Trace[_R]]: ...

    @overload
    def __call__(self, fn: Callable[_P, _R]) -> SyncVersionedFunction[_P, Trace[_R]]: ...

    def __call__(
        self,
        fn: TraceDecoratedFunctionWithContext[_P, Coroutine[Any, Any, _R]]
        | TraceDecoratedFunctionWithContext[_P, _R]
        | Callable[_P, _R]
        | Callable[_P, Coroutine[Any, Any, _R]],
    ) -> AsyncVersionedFunction[_P, Trace[_R]] | SyncVersionedFunction[_P, Trace[_R]]:
        """Protocol `call` definition for `VersionedFunction` decorator return type."""
        ...


_TraceAttribute: TypeAlias = dict[str, AttributeValue]


class _ResultHolder:
    """A class to hold the result of a function call."""

    def __init__(self) -> None:
        self.result = None

    def set_result(self, result: Any) -> None:
        """Set the result attribute."""
        self.result: Any = result


_trace_context: ContextVar[MappingProxyType] = ContextVar("_trace_context", default=MappingProxyType({}))


def _get_trace_context() -> dict[str, Any]:
    return dict(_trace_context.get())


def _set_trace_context(trace_ctx: dict[str, Any]) -> None:
    _trace_context.set(MappingProxyType(trace_ctx.copy()))


@contextmanager
def _set_span_attributes(
    span: Span,
    span_attribute: _TraceAttribute,
    is_async: bool,
    function: FunctionPublic | None,
    decorator_tags: list[str] | None = None,
    serializers: SerializerMap | None = None,
) -> Generator[_ResultHolder, None, None]:
    """Set the attributes on the span."""
    settings = get_settings()
    span_attribute["lilypad.project_uuid"] = settings.project_id if settings.project_id else ""
    span_attribute["lilypad.type"] = trace_type = _get_trace_type(function)
    span_attribute["lilypad.is_async"] = is_async
    if decorator_tags is not None:
        span_attribute["lilypad.trace.tags"] = decorator_tags
    if function:
        function_uuid = function.uuid
        span_attribute[f"lilypad.{trace_type}.signature"] = function.signature
        span_attribute[f"lilypad.{trace_type}.code"] = function.code
    else:
        function_uuid = ""
    span_attribute["lilypad.function.uuid"] = function_uuid
    span.opentelemetry_span.set_attributes(span_attribute)
    result_holder = _ResultHolder()
    yield result_holder
    original_output = result_holder.result
    span.opentelemetry_span.set_attribute(
        f"lilypad.{trace_type}.output", "" if original_output is None else to_text(original_output, serializers)
    )


def _construct_trace_attributes(
    trace_type: str,
    arg_types: dict[str, str],
    arg_values: dict[str, Any],
    serializers: SerializerMap,
) -> dict[str, AttributeValue]:
    jsonable_arg_values = {}
    for arg_name, arg_value in arg_values.items():
        try:
            serialized_arg_value = fast_jsonable(arg_value, custom_serializers=serializers)
        except (TypeError, ValueError, orjson.JSONEncodeError):
            serialized_arg_value = "could not serialize"
        jsonable_arg_values[arg_name] = serialized_arg_value
    return {
        f"lilypad.{trace_type}.arg_types": json_dumps(arg_types),
        f"lilypad.{trace_type}.arg_values": json_dumps(jsonable_arg_values),
    }


_SANDBOX_CUSTOM_RESULT = {
    "result": "result",
    "trace_context": "_get_trace_context()",
}
_SANDBOX_PRE_ACTIONS = [
    "lilypad.configure(log_handlers=[logging.StreamHandler(sys.stderr)])",
]
_SANDBOX_AFTER_ACTIONS = [
    "result = result.response if isinstance(result, AsyncTrace | Trace) else result",
    "with suppress(ImportError): from mirascope.core import BaseCallResponse",
    "result = getattr(result, 'content', result) if (mirascope_response := locals().get('BaseCallResponse')) else result",
]
_SANDBOX_EXTRA_IMPORT = [
    f"from {TRACE_MODULE_NAME} import _get_trace_context, AsyncTrace, Trace",
    "import lilypad",
    "import sys",
    "import logging",
    "from contextlib import suppress",
]


@overload
def trace(
    name: str | None = None,
    *,
    versioning: None = None,
    mode: None = None,
    tags: list[str] | None = None,
    serializers: SerializerMap | None = None,
) -> TraceDecorator: ...


@overload
def trace(
    name: str | None = None,
    *,
    versioning: Literal["automatic"],
    mode: None = None,
    tags: list[str] | None = None,
    serializers: SerializerMap | None = None,
) -> VersionedFunctionTraceDecorator: ...


@overload
def trace(
    name: str | None = None,
    *,
    versioning: None,
    mode: Literal["wrap"],
    tags: list[str] | None = None,
    serializers: SerializerMap | None = None,
) -> WrappedTraceDecorator: ...


@overload
def trace(
    name: str | None = None,
    *,
    versioning: Literal["automatic"],
    mode: Literal["wrap"],
    tags: list[str] | None = None,
    serializers: SerializerMap | None = None,
) -> WrappedVersionedFunctionTraceDecorator: ...


def trace(
    name: str | None = None,
    *,
    versioning: Literal["automatic"] | None = None,
    mode: Literal["wrap"] | None = None,
    tags: list[str] | None = None,
    serializers: SerializerMap | None = None,
) -> TraceDecorator | VersionedFunctionTraceDecorator:
    """The tracing LLM generations.

    The decorated function will trace and log automatically.
    If mode="wrap" is set, the function will return a Trace[_R] object with a 'response' property containing the original function's response and an 'annotate' method.
    """

    decorator_tags = sorted(list(set(tags))) if tags else None

    @overload
    def decorator(
        fn: TraceDecoratedFunctionWithContext[_P, Coroutine[Any, Any, _R]],
    ) -> Callable[_P, Coroutine[Any, Any, _R]]: ...

    @overload
    def decorator(fn: TraceDecoratedFunctionWithContext[_P, _R]) -> Callable[_P, _R]: ...

    @overload
    def decorator(fn: Callable[_P, Coroutine[Any, Any, _R]]) -> Callable[_P, Coroutine[Any, Any, _R]]: ...

    @overload
    def decorator(fn: Callable[_P, _R]) -> Callable[_P, _R]: ...

    def decorator(
        fn: TraceDecoratedFunctionWithContext[_P, Coroutine[Any, Any, _R]]
        | TraceDecoratedFunctionWithContext[_P, _R]
        | Callable[_P, _R]
        | Callable[_P, Coroutine[Any, Any, _R]],
    ) -> Callable[_P, _R] | Callable[_P, Coroutine[Any, Any, _R]]:
        is_mirascope_call = hasattr(fn, "__mirascope_call__")
        prompt_template = (
            fn._prompt_template if hasattr(fn, "_prompt_template") else ""  # pyright: ignore[reportFunctionMemberAccess]
        )

        if _RECORDING_ENABLED and versioning == "automatic":
            _register_decorated_function(
                TRACE_MODULE_NAME, fn, Closure.from_fn(fn).name, {"mode": mode, "tags": decorator_tags}
            )

        local_serializers = serializers or {}

        settings = get_settings()

        signature = get_signature(fn)

        if name is None:
            trace_name = get_qualified_name(fn)
        else:
            trace_name = name
        if fn_is_async(fn):

            @call_safely(fn, catch=NETWORK_EXCEPTIONS)
            @wraps(fn)
            async def inner_async(*args: _P.args, **kwargs: _P.kwargs) -> _R:
                with Span(trace_name) as span:
                    final_args = args
                    final_kwargs = kwargs
                    needs_trace_ctx = "trace_ctx" in signature.parameters
                    has_user_provided_trace_ctx = False
                    try:
                        bound_call_args = signature.bind(*args, **kwargs)
                        has_user_provided_trace_ctx = "trace_ctx" in bound_call_args.arguments
                    except TypeError:
                        pass
                    if needs_trace_ctx and not has_user_provided_trace_ctx:
                        final_args = tuple((span, *args))
                    arg_types, arg_values = inspect_arguments(fn, *final_args, **final_kwargs)
                    arg_values.pop("trace_ctx", None)
                    arg_types.pop("trace_ctx", None)

                    async_lilypad_client = get_async_client(api_key=settings.api_key)

                    closure = Closure.from_fn(fn)

                    @call_safely(fn, catch=NETWORK_EXCEPTIONS)
                    @wraps(fn)
                    async def get_or_create_function_async() -> FunctionPublic | None:
                        try:
                            return await get_function_by_hash_async(
                                project_uuid=settings.project_id, function_hash=closure.hash
                            )
                        except NotFoundError:
                            return await async_lilypad_client.projects.functions.create(
                                path_project_uuid=settings.project_id,
                                code=closure.code,
                                hash=closure.hash,
                                name=closure.name,
                                signature=closure.signature,
                                arg_types=arg_types,
                                dependencies=closure.dependencies,
                                is_versioned=True,
                                prompt_template=prompt_template,
                            )

                    if versioning == "automatic":
                        function = await get_or_create_function_async()
                    else:
                        function = None

                    function_uuid = function.uuid if function else None

                    trace_attribute = _construct_trace_attributes(
                        trace_type=_get_trace_type(function),
                        arg_types=arg_types,
                        arg_values=arg_values,
                        serializers=local_serializers,
                    )

                    if is_mirascope_call:
                        decorator_inner = create_mirascope_middleware(
                            function,
                            arg_types,
                            arg_values,
                            True,
                            prompt_template,
                            settings.project_id,
                            current_span=span.opentelemetry_span,
                        )
                        output = await decorator_inner(fn)(*final_args, **final_kwargs)
                    else:
                        with _set_span_attributes(
                            span, trace_attribute, is_async=True, function=function, serializers=local_serializers
                        ) as result_holder:
                            output = await fn(*final_args, **final_kwargs)
                            result_holder.set_result(output)
                    span_id = span.span_id
                    _set_trace_context({"span_id": span_id, "function_uuid": function_uuid})

                if mode == "wrap":
                    return AsyncTrace(response=output, span_id=span_id, function_uuid=function_uuid)
                return output  # pyright: ignore [reportReturnType]

            if versioning is None:
                return inner_async

            function_name = get_qualified_name(fn)

            async def _specific_function_version_async(
                forced_version: int,
                sandbox: SandboxRunner | None = None,
            ) -> Callable[_P, _R]:
                try:
                    versioned_function = await get_function_by_version_async(
                        version_num=forced_version,
                        project_uuid=settings.project_id,
                        function_name=function_name,
                    )
                    versioned_function_closure = get_cached_closure(versioned_function)
                except Exception as e:
                    raise RemoteFunctionError(f"Failed to retrieve function {fn.__name__}: {e}")

                if sandbox is None:
                    sandbox = SubprocessSandboxRunner(os.environ.copy())

                @call_safely(fn, catch=NETWORK_EXCEPTIONS)  # pyright: ignore [reportArgumentType]
                @wraps(fn)
                def _inner_async(*args: _P.args, **kwargs: _P.kwargs) -> _R:
                    result = sandbox.execute_function(
                        versioned_function_closure,
                        *args,
                        custom_result=_SANDBOX_CUSTOM_RESULT,
                        pre_actions=_SANDBOX_PRE_ACTIONS,
                        after_actions=_SANDBOX_AFTER_ACTIONS,
                        extra_imports=_SANDBOX_EXTRA_IMPORT,
                        **kwargs,
                    )
                    if mode == "wrap":
                        return AsyncTrace(
                            response=result["result"],
                            span_id=result["trace_context"]["span_id"],
                            function_uuid=result["trace_context"]["function_uuid"],
                        )
                    return result["result"]

                return _inner_async

            inner_async.version = _specific_function_version_async  # pyright: ignore [reportAttributeAccessIssue, reportFunctionMemberAccess]

            async def _deployed_version_async(
                *args: _P.args,
                sandbox: SandboxRunner | None = None,
                ttl: float | None = None,
                force_refresh: bool = False,
                **kwargs: _P.kwargs,
            ) -> _R:
                try:
                    deployed_function = await get_deployed_function_async(
                        project_uuid=settings.project_id,
                        ttl=ttl,
                        force_refresh=force_refresh,
                        function_name=function_name,
                    )
                    deployed_function_closure = get_cached_closure(deployed_function)
                except Exception as e:
                    raise RemoteFunctionError(f"Failed to retrieve function {fn.__name__}: {e}")

                if sandbox is None:
                    sandbox = SubprocessSandboxRunner(os.environ.copy())

                result = sandbox.execute_function(
                    deployed_function_closure,
                    *args,
                    custom_result=_SANDBOX_CUSTOM_RESULT,
                    pre_actions=_SANDBOX_PRE_ACTIONS,
                    after_actions=_SANDBOX_AFTER_ACTIONS,
                    extra_imports=_SANDBOX_EXTRA_IMPORT,
                    **kwargs,
                )
                if mode == "wrap":
                    return AsyncTrace(
                        response=result["result"],
                        span_id=result["trace_context"]["span_id"],
                        function_uuid=result["trace_context"]["function_uuid"],
                    )
                return result["result"]

            inner_async.remote = _deployed_version_async
            return inner_async
        else:

            @call_safely(fn, catch=NETWORK_EXCEPTIONS)
            @wraps(fn)
            def inner(*args: _P.args, **kwargs: _P.kwargs) -> _R:
                with Span(trace_name) as span:
                    final_args = args
                    final_kwargs = kwargs
                    needs_trace_ctx = "trace_ctx" in signature.parameters
                    has_user_provided_trace_ctx = False
                    try:
                        bound_call_args = signature.bind(*args, **kwargs)
                        has_user_provided_trace_ctx = "trace_ctx" in bound_call_args.arguments
                    except TypeError:
                        pass

                    if needs_trace_ctx and not has_user_provided_trace_ctx:
                        final_args = tuple((span, *args))
                    arg_types, arg_values = inspect_arguments(fn, *final_args, **final_kwargs)
                    arg_values.pop("trace_ctx", None)
                    arg_types.pop("trace_ctx", None)

                    lilypad_client = get_sync_client(api_key=settings.api_key)

                    closure = Closure.from_fn(fn)

                    @call_safely(fn, catch=NETWORK_EXCEPTIONS)
                    @wraps(fn)
                    def get_or_create_function_sync() -> FunctionPublic | None:
                        try:
                            return get_function_by_hash_sync(
                                project_uuid=settings.project_id, function_hash=closure.hash
                            )
                        except NotFoundError:
                            return lilypad_client.projects.functions.create(
                                path_project_uuid=settings.project_id,
                                code=closure.code,
                                hash=closure.hash,
                                name=closure.name,
                                signature=closure.signature,
                                arg_types=arg_types,
                                dependencies=closure.dependencies,
                                is_versioned=True,
                                prompt_template=prompt_template,
                            )

                    if versioning == "automatic":
                        function = get_or_create_function_sync()
                    else:
                        function = None

                    function_uuid = function.uuid if function else None

                    trace_attribute = _construct_trace_attributes(
                        trace_type=_get_trace_type(function),
                        arg_types=arg_types,
                        arg_values=arg_values,
                        serializers=local_serializers,
                    )

                    if is_mirascope_call:
                        decorator_inner = create_mirascope_middleware(
                            function,
                            arg_types,
                            arg_values,
                            False,
                            prompt_template,
                            settings.project_id,
                            current_span=span.opentelemetry_span,
                            decorator_tags=decorator_tags,
                        )
                        output = decorator_inner(fn)(*final_args, **final_kwargs)
                    else:
                        with _set_span_attributes(
                            span,
                            trace_attribute,
                            is_async=False,
                            function=function,
                            decorator_tags=decorator_tags,
                            serializers=local_serializers,
                        ) as result_holder:
                            output = fn(*final_args, **final_kwargs)
                            result_holder.set_result(output)
                    span_id = span.span_id
                    _set_trace_context({"span_id": span_id, "function_uuid": function_uuid})

                if mode == "wrap":
                    return Trace(response=output, span_id=span_id, function_uuid=function_uuid)
                return output  # pyright: ignore [reportReturnType]

            if versioning is None:
                return inner  # pyright: ignore [reportReturnType]

            function_name = get_qualified_name(fn)

            def _specific_function_version(
                forced_version: int,
                sandbox: SandboxRunner | None = None,
            ) -> Callable[_P, _R]:
                try:
                    versioned_function = get_function_by_version_sync(
                        version_num=forced_version,
                        project_uuid=settings.project_id,
                        function_name=function_name,
                    )
                    versioned_function_closure = get_cached_closure(versioned_function)
                except Exception as e:
                    raise RemoteFunctionError(f"Failed to retrieve function {fn.__name__}: {e}")

                if sandbox is None:
                    sandbox = SubprocessSandboxRunner(os.environ.copy())

                @call_safely(fn, catch=NETWORK_EXCEPTIONS)  # pyright: ignore [reportArgumentType]
                @wraps(fn)
                def _inner(*args: _P.args, **kwargs: _P.kwargs) -> _R:
                    result = sandbox.execute_function(
                        versioned_function_closure,
                        *args,
                        custom_result=_SANDBOX_CUSTOM_RESULT,
                        pre_actions=_SANDBOX_PRE_ACTIONS,
                        after_actions=_SANDBOX_AFTER_ACTIONS,
                        extra_imports=_SANDBOX_EXTRA_IMPORT,
                        **kwargs,
                    )
                    if mode == "wrap":
                        return Trace(
                            response=result["result"],
                            span_id=result["trace_context"]["span_id"],
                            function_uuid=result["trace_context"]["function_uuid"],
                        )
                    return result["result"]

                return _inner

            inner.version = _specific_function_version  # pyright: ignore [reportAttributeAccessIssue, reportFunctionMemberAccess]

            def _deployed_version(
                *args: _P.args,
                sandbox: SandboxRunner | None = None,
                ttl: float | None = None,
                force_refresh: bool = False,
                **kwargs: _P.kwargs,
            ) -> _R:
                try:
                    deployed_function = get_deployed_function_sync(
                        project_uuid=settings.project_id,
                        ttl=ttl,
                        force_refresh=force_refresh,
                        function_name=function_name,
                    )
                    deployed_function_closure = get_cached_closure(deployed_function)
                except Exception as e:
                    raise RemoteFunctionError(f"Failed to retrieve function {fn.__name__}: {e}")

                if sandbox is None:
                    sandbox = SubprocessSandboxRunner(os.environ.copy())

                result = sandbox.execute_function(
                    deployed_function_closure,
                    *args,
                    custom_result=_SANDBOX_CUSTOM_RESULT,
                    pre_actions=_SANDBOX_PRE_ACTIONS,
                    after_actions=_SANDBOX_AFTER_ACTIONS,
                    extra_imports=_SANDBOX_EXTRA_IMPORT,
                    **kwargs,
                )
                if mode == "wrap":
                    return Trace(
                        response=result["result"],
                        span_id=result["trace_context"]["span_id"],
                        function_uuid=result["trace_context"]["function_uuid"],
                    )
                return result["result"]

            inner.remote = _deployed_version
            return inner

    return decorator
